@@ -1,34 +1,40 @@
 """Extensions to unittest for web frameworks.
 
-Use the WebCase.getPage method to request a page from your HTTP server.
+Use the :py:meth:`WebCase.getPage` method to request a page
+from your HTTP server.
+
 Framework Integration
 =====================
 If you have control over your server process, you can handle errors
 in the server-side of the HTTP conversation a bit better. You must run
-both the client (your WebCase tests) and the server in the same process
-(but in separate threads, obviously).
+both the client (your :py:class:`WebCase` tests) and the server in the
+same process (but in separate threads, obviously).
 When an error occurs in the framework, call server_error. It will print
 the traceback to stdout, and keep any assertions you have from running
 (the assumption is that, if the server errors, the page output will not
 be of further significance to your tests).
 """
 
+from __future__ import absolute_import, division, print_function
+__metaclass__ = type
+
 import pprint
 import re
 import socket
 import sys
 import time
 import traceback
-import types
 import os
 import json
 import unittest
-from imp import reload
+import warnings
+import functools
 
-from six.moves import range, http_client, map, urllib_parse
+from six.moves import http_client, map, urllib_parse
 import six
 
 from more_itertools.more import always_iterable
+import jaraco.functools
 
 
 def interface(host):
@@ -46,109 +52,11 @@ def interface(host):
     return host
 
 
-class TerseTestResult(unittest._TextTestResult):
-
-    def printErrors(self):
-        # Overridden to avoid unnecessary empty line
-        if self.errors or self.failures:
-            if self.dots or self.showAll:
-                self.stream.writeln()
-            self.printErrorList('ERROR', self.errors)
-            self.printErrorList('FAIL', self.failures)
-
-
-class TerseTestRunner(unittest.TextTestRunner):
-    """A test runner class that displays results in textual form."""
-
-    def _makeResult(self):
-        return TerseTestResult(self.stream, self.descriptions, self.verbosity)
-
-    def run(self, test):
-        """Run the given test case or test suite."""
-        # Overridden to remove unnecessary empty lines and separators
-        result = self._makeResult()
-        test(result)
-        result.printErrors()
-        if not result.wasSuccessful():
-            self.stream.write('FAILED (')
-            failed, errored = list(map(len, (result.failures, result.errors)))
-            if failed:
-                self.stream.write('failures=%d' % failed)
-            if errored:
-                if failed:
-                    self.stream.write(', ')
-                self.stream.write('errors=%d' % errored)
-            self.stream.writeln(')')
-        return result
-
-
-class ReloadingTestLoader(unittest.TestLoader):
-
-    def loadTestsFromName(self, name, module=None):
-        """Return a suite of all tests cases given a string specifier.
-
-        The name may resolve either to a module, a test case class, a
-        test method within a test case class, or a callable object which
-        returns a TestCase or TestSuite instance.
-        The method optionally resolves the names relative to a given module.
-        """
-        parts = name.split('.')
-        unused_parts = []
-        if module is None:
-            if not parts:
-                raise ValueError('incomplete test name: %s' % name)
-            else:
-                parts_copy = parts[:]
-                while parts_copy:
-                    target = '.'.join(parts_copy)
-                    if target in sys.modules:
-                        module = reload(sys.modules[target])
-                        parts = unused_parts
-                        break
-                    else:
-                        try:
-                            module = __import__(target)
-                            parts = unused_parts
-                            break
-                        except ImportError:
-                            unused_parts.insert(0, parts_copy[-1])
-                            del parts_copy[-1]
-                            if not parts_copy:
-                                raise
-                parts = parts[1:]
-        obj = module
-        for part in parts:
-            obj = getattr(obj, part)
-
-        if isinstance(obj, types.ModuleType):
-            return self.loadTestsFromModule(obj)
-        elif (
-                (
-                    (six.PY3 and isinstance(obj, type)) or
-                    isinstance(obj, (type, types.ClassType))
-                ) and
-                issubclass(obj, unittest.TestCase)):
-            return self.loadTestsFromTestCase(obj)
-        elif isinstance(obj, types.UnboundMethodType):
-            if six.PY3:
-                return obj.__self__.__class__(obj.__name__)
-            else:
-                return obj.im_class(obj.__name__)
-        elif hasattr(obj, '__call__'):
-            test = obj()
-            if not isinstance(test, unittest.TestCase) and \
-               not isinstance(test, unittest.TestSuite):
-                raise ValueError('calling %s returned %s, '
-                                 'not a test' % (obj, test))
-            return test
-        else:
-            raise ValueError('do not know how to make test from: %s' % obj)
-
-
 try:
     # Jython support
     if sys.platform[:4] == 'java':
         def getchar():
+            """Get a key press."""
             # Hopefully this is enough
             return sys.stdin.read(1)
     else:
@@ -156,6 +64,7 @@ try:
         import msvcrt
 
         def getchar():
+            """Get a key press."""
             return msvcrt.getch()
 except ImportError:
     # Unix getchr
@@ -163,6 +72,7 @@ except ImportError:
     import termios
 
     def getchar():
+        """Get a key press."""
         fd = sys.stdin.fileno()
         old_settings = termios.tcgetattr(fd)
         try:
@@ -174,19 +84,25 @@ except ImportError:
 
 
 # from jaraco.properties
-class NonDataProperty(object):
+class NonDataProperty:
+    """Non-data property decorator."""
+
     def __init__(self, fget):
+        """Initialize a non-data property."""
         assert fget is not None, 'fget cannot be none'
         assert callable(fget), 'fget must be callable'
         self.fget = fget
 
     def __get__(self, obj, objtype=None):
+        """Return a class property."""
         if obj is None:
             return self
         return self.fget(obj)
 
 
 class WebCase(unittest.TestCase):
+    """Helper web test suite base."""
+
     HOST = '127.0.0.1'
     PORT = 8000
     HTTP_CONN = http_client.HTTPConnection
@@ -194,6 +110,7 @@ class WebCase(unittest.TestCase):
 
     scheme = 'http'
     url = None
+    ssl_context = None
 
     status = None
     headers = None
@@ -205,9 +122,9 @@ class WebCase(unittest.TestCase):
 
     @property
     def _Conn(self):
-        """
-        Return http.client.HTTPConnection or HTTPSConnection
-        based on self.scheme.
+        """Return HTTPConnection or HTTPSConnection based on self.scheme.
+
+        * from http.client.
         """
         cls_name = '{scheme}Connection'.format(scheme=self.scheme.upper())
         return getattr(http_client, cls_name)
@@ -224,10 +141,10 @@ class WebCase(unittest.TestCase):
         """Make our HTTP_CONN persistent (or not).
 
         If the 'on' argument is True (the default), then self.HTTP_CONN
-        will be set to an instance of HTTPConnection (or HTTPS
-        if self.scheme is "https"). This will then persist across requests.
-        We only allow for a single open connection, so if you call this
-        and we currently have an open connection, it will be closed.
+        will be set to an instance of HTTP(S)?Connection
+        to persist across requests.
+        As this class only allows for a single open connection, if
+        self already has an open connection, it will be closed.
         """
         try:
             self.HTTP_CONN.close()
@@ -242,6 +159,7 @@ class WebCase(unittest.TestCase):
 
     @property
     def persistent(self):
+        """Presence of the persistent HTTP connection."""
         return hasattr(self.HTTP_CONN, '__class__')
 
     @persistent.setter
@@ -256,8 +174,10 @@ class WebCase(unittest.TestCase):
         """
         return interface(self.HOST)
 
-    def getPage(self, url, headers=None, method='GET', body=None,
-                protocol=None, raise_subcls=None):
+    def getPage(
+        self, url, headers=None, method='GET', body=None,
+        protocol=None, raise_subcls=(),
+    ):
         """Open the url with debugging support. Return status, headers, body.
 
         url should be the identifier passed to the server, typically a
@@ -266,19 +186,16 @@ class WebCase(unittest.TestCase):
         enabled in the server.
 
         If the application under test generates absolute URIs, be sure
-        to wrap them first with strip_netloc::
+        to wrap them first with :py:func:`strip_netloc`::
 
-            class MyAppWebCase(WebCase):
-                def getPage(url, *args, **kwargs):
-                    super(MyAppWebCase, self).getPage(
-                        cheroot.test.webtest.strip_netloc(url),
-                        *args, **kwargs
-                    )
+            >>> class MyAppWebCase(WebCase):
+            ...     def getPage(url, *args, **kwargs):
+            ...         super(MyAppWebCase, self).getPage(
+            ...             cheroot.test.webtest.strip_netloc(url),
+            ...             *args, **kwargs
+            ...         )
 
-        `raise_subcls` must be a tuple with the exceptions classes
-        or a single exception class that are not going to be considered
-        a socket.error regardless that they were are subclass of a
-        socket.error and therefore not considered for a connection retry.
+        `raise_subcls` is passed through to :py:func:`openURL`.
         """
         ServerError.on = False
 
@@ -287,18 +204,26 @@ class WebCase(unittest.TestCase):
         if isinstance(body, six.text_type):
             body = body.encode('utf-8')
 
+        # for compatibility, support raise_subcls is None
+        raise_subcls = raise_subcls or ()
+
         self.url = url
         self.time = None
         start = time.time()
-        result = openURL(url, headers, method, body, self.HOST, self.PORT,
-                         self.HTTP_CONN, protocol or self.PROTOCOL,
-                         raise_subcls)
+        result = openURL(
+            url, headers, method, body, self.HOST, self.PORT,
+            self.HTTP_CONN, protocol or self.PROTOCOL,
+            raise_subcls=raise_subcls,
+            ssl_context=self.ssl_context,
+        )
         self.time = time.time() - start
         self.status, self.headers, self.body = result
 
         # Build a list of request cookies from the previous response cookies.
-        self.cookies = [('Cookie', v) for k, v in self.headers
-                        if k.lower() == 'set-cookie']
+        self.cookies = [
+            ('Cookie', v) for k, v in self.headers
+            if k.lower() == 'set-cookie'
+        ]
 
         if ServerError.on:
             raise ServerError()
@@ -313,7 +238,14 @@ class WebCase(unittest.TestCase):
         False or 1 or 0.
         """
         env_str = os.environ.get('WEBTEST_INTERACTIVE', 'True')
-        return bool(json.loads(env_str.lower()))
+        is_interactive = bool(json.loads(env_str.lower()))
+        if is_interactive:
+            warnings.warn(
+                'Interactive test failure interceptor support via '
+                'WEBTEST_INTERACTIVE environment variable is deprecated.',
+                DeprecationWarning,
+            )
+        return is_interactive
 
     console_height = 30
 
@@ -324,9 +256,11 @@ class WebCase(unittest.TestCase):
         if not self.interactive:
             raise self.failureException(msg)
 
-        p = ('    Show: '
-             '[B]ody [H]eaders [S]tatus [U]RL; '
-             '[I]gnore, [R]aise, or sys.e[X]it >> ')
+        p = (
+            '    Show: '
+            '[B]ody [H]eaders [S]tatus [U]RL; '
+            '[I]gnore, [R]aise, or sys.e[X]it >> '
+        )
         sys.stdout.write(p)
         sys.stdout.flush()
         while True:
@@ -359,18 +293,17 @@ class WebCase(unittest.TestCase):
             elif i == 'R':
                 raise self.failureException(msg)
             elif i == 'X':
-                self.exit()
+                sys.exit()
             sys.stdout.write(p)
             sys.stdout.flush()
 
-    def exit(self):
-        sys.exit()
-
     @property
-    def status_code(self):
+    def status_code(self):  # noqa: D401; irrelevant for properties
+        """Integer HTTP status code."""
         return int(self.status[:3])
 
     def status_matches(self, expected):
+        """Check whether actual status matches expected."""
         actual = (
             self.status_code
             if isinstance(expected, int) else
@@ -456,7 +389,8 @@ class WebCase(unittest.TestCase):
         if value != self.body:
             if msg is None:
                 msg = 'expected body:\n%r\n\nactual body:\n%r' % (
-                    value, self.body)
+                    value, self.body,
+                )
             self._handlewebError(msg)
 
     def assertInBody(self, value, msg=None):
@@ -517,7 +451,8 @@ def cleanHeaders(headers, method, body, host, port):
                 break
         if not found:
             headers.append(
-                ('Content-Type', 'application/x-www-form-urlencoded'))
+                ('Content-Type', 'application/x-www-form-urlencoded'),
+            )
             headers.append(('Content-Length', str(len(body or ''))))
 
     return headers
@@ -525,91 +460,102 @@ def cleanHeaders(headers, method, body, host, port):
 
 def shb(response):
     """Return status, headers, body the way we like from a response."""
-    if six.PY3:
-        h = response.getheaders()
-    else:
-        h = []
-        key, value = None, None
-        for line in response.msg.headers:
-            if line:
-                if line[0] in ' \t':
-                    value += line.strip()
-                else:
-                    if key and value:
-                        h.append((key, value))
-                    key, value = line.split(':', 1)
-                    key = key.strip()
-                    value = value.strip()
-        if key and value:
-            h.append((key, value))
+    resp_status_line = '%s %s' % (response.status, response.reason)
 
-    return '%s %s' % (response.status, response.reason), h, response.read()
+    if not six.PY2:
+        return resp_status_line, response.getheaders(), response.read()
+
+    h = []
+    key, value = None, None
+    for line in response.msg.headers:
+        if line:
+            if line[0] in ' \t':
+                value += line.strip()
+            else:
+                if key and value:
+                    h.append((key, value))
+                key, value = line.split(':', 1)
+                key = key.strip()
+                value = value.strip()
+    if key and value:
+        h.append((key, value))
+
+    return resp_status_line, h, response.read()
 
 
-def openURL(url, headers=None, method='GET', body=None,
-            host='127.0.0.1', port=8000, http_conn=http_client.HTTPConnection,
-            protocol='HTTP/1.1', raise_subcls=None):
+# def openURL(*args, raise_subcls=(), **kwargs):
+# py27 compatible signature:
+def openURL(*args, **kwargs):
     """
-    Open the given HTTP resource and return status, headers, and body.
+    Open a URL, retrying when it fails.
 
-    `raise_subcls` must be a tuple with the exceptions classes
-    or a single exception class that are not going to be considered
-    a socket.error regardless that they were are subclass of a
-    socket.error and therefore not considered for a connection retry.
+    Specify `raise_subcls` (class or tuple of classes) to exclude
+    those socket.error subclasses from being suppressed and retried.
     """
+    raise_subcls = kwargs.pop('raise_subcls', ())
+    opener = functools.partial(_open_url_once, *args, **kwargs)
+
+    def on_exception():
+        type_, exc = sys.exc_info()[:2]
+        if isinstance(exc, raise_subcls):
+            raise
+        time.sleep(0.5)
+
+    # Try up to 10 times
+    return jaraco.functools.retry_call(
+        opener,
+        retries=9,
+        cleanup=on_exception,
+        trap=socket.error,
+    )
+
+
+def _open_url_once(
+    url, headers=None, method='GET', body=None,
+    host='127.0.0.1', port=8000, http_conn=http_client.HTTPConnection,
+    protocol='HTTP/1.1', ssl_context=None,
+):
+    """Open the given HTTP resource and return status, headers, and body."""
     headers = cleanHeaders(headers, method, body, host, port)
 
-    # Trying 10 times is simply in case of socket errors.
-    # Normal case--it should run once.
-    for trial in range(10):
-        try:
-            # Allow http_conn to be a class or an instance
-            if hasattr(http_conn, 'host'):
-                conn = http_conn
-            else:
-                conn = http_conn(interface(host), port)
-
-            conn._http_vsn_str = protocol
-            conn._http_vsn = int(''.join([x for x in protocol if x.isdigit()]))
-
-            if six.PY3 and isinstance(url, bytes):
-                url = url.decode()
-            conn.putrequest(method.upper(), url, skip_host=True,
-                            skip_accept_encoding=True)
-
-            for key, value in headers:
-                conn.putheader(key, value.encode('Latin-1'))
-            conn.endheaders()
-
-            if body is not None:
-                conn.send(body)
-
-            # Handle response
-            response = conn.getresponse()
-
-            s, h, b = shb(response)
-
-            if not hasattr(http_conn, 'host'):
-                # We made our own conn instance. Close it.
-                conn.close()
-
-            return s, h, b
-        except socket.error as e:
-            if raise_subcls is not None and isinstance(e, raise_subcls):
-                raise
-            else:
-                time.sleep(0.5)
-                if trial == 9:
-                    raise
+    # Allow http_conn to be a class or an instance
+    if hasattr(http_conn, 'host'):
+        conn = http_conn
+    else:
+        kw = {}
+        if ssl_context:
+            kw['context'] = ssl_context
+        conn = http_conn(interface(host), port, **kw)
+    conn._http_vsn_str = protocol
+    conn._http_vsn = int(''.join([x for x in protocol if x.isdigit()]))
+    if not six.PY2 and isinstance(url, bytes):
+        url = url.decode()
+    conn.putrequest(
+        method.upper(), url, skip_host=True,
+        skip_accept_encoding=True,
+    )
+    for key, value in headers:
+        conn.putheader(key, value.encode('Latin-1'))
+    conn.endheaders()
+    if body is not None:
+        conn.send(body)
+    # Handle response
+    response = conn.getresponse()
+    s, h, b = shb(response)
+    if not hasattr(http_conn, 'host'):
+        # We made our own conn instance. Close it.
+        conn.close()
+    return s, h, b
 
 
 def strip_netloc(url):
-    """
+    """Return absolute-URI path from URL.
+
     Strip the scheme and host from the URL, returning the
     server-absolute portion.
 
     Useful for wrapping an absolute-URI for which only the
-    path is expected (such as in calls to getPage).
+    path is expected (such as in calls to :py:meth:`WebCase.getPage`).
 
     >>> strip_netloc('https://google.com/foo/bar?bing#baz')
     '/foo/bar?bing'
@@ -637,6 +583,8 @@ ignore_all = False
 
 
 class ServerError(Exception):
+    """Exception for signalling server error."""
+
     on = False
 
 
